/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core;

import org.apache.commons.lang.text.StrBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Map;

/**
 * {@link CapabilitiesHolder} default implementation.
 * <p>
 * This is the default implementation of CapabilitiesHolder, it use a Map to
 * store capabilities and a CapabilitiesLoader implementation to lazily load
 * capabilities values.
 * </p>
 * <p/>
 * <p>
 * It is very efficient if you need getting more then one capability, and if
 * you want getting capability from JSTL, because you can't use
 * getCapability("name") from there but you can grab capability by doing
 * ${device.capabilities.name}.
 * </p>
 *
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * @version $Id: DefaultCapabilitiesHolder.java 649 2011-05-20 18:28:18Z ermanno.franco $
 */
public class DefaultCapabilitiesHolder implements CapabilitiesHolder,
        Serializable {

    /**
     * serial
     */
    private static final long serialVersionUID = 100L;

    private final Logger logger = LoggerFactory.getLogger(DefaultCapabilitiesHolder.class);


    /**
     * Capability loader will not serialized because it can referencing the
     * Model
     */
    private transient CapabilitiesLoader capabilitiesLoader;

    /**
     * The map where the capabilities are stored
     */
    private Map capabilities;


    public DefaultCapabilitiesHolder(CapabilitiesLoader capabilitiesLoader) {
        this.capabilitiesLoader = capabilitiesLoader;
    }

    public String getCapabilityValue(String capabilityName) throws CapabilityNotDefinedException {

        String value = (String) getAllCapabilities().get(capabilityName);

        if (logger.isTraceEnabled()) {
            StrBuilder msgBuilder = new StrBuilder();
            msgBuilder.append("Get capability: ").append(capabilityName).append(" = ").append(value);

            logger.trace(msgBuilder.toString());
        }

        if (value == null) {
            throw new CapabilityNotDefinedException(capabilityName);
        } else {
            return value;
        }
    }

    public Map getAllCapabilities() {
        if (this.capabilities == null) {

            if (capabilitiesLoader == null) {
                throw new IllegalStateException(
                        "The device must be initialized before serialization");
            } else {
                logger.debug("Initializing capabilities");
                this.capabilities = capabilitiesLoader.loadCapabilities();
            }
        }

        assert capabilities != null;

        return capabilities;
    }

    // Serialization ******************************************************

    private void writeObject(ObjectOutputStream out) throws IOException {

        if (this.capabilities == null) {
            this.capabilities = capabilitiesLoader.loadCapabilities();
        }

        out.defaultWriteObject();
	}
	

}
