/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core;

import net.sourceforge.wurfl.core.resource.ModelDevice;
import net.sourceforge.wurfl.core.utils.StringUtils;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

/**
 * This is a default Device implementation.
 * <p/>
 * <p>
 * It uses a CapabilitiesHolder to retrieve capabilities values and
 * MarkupResolver to resolve MarkUp.
 * </p>
 *
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * @version $Id: DefaultDevice.java 668 2011-06-14 07:07:44Z ermanno.franco $
 */
public class DefaultDevice implements Device, Serializable {

    /**
     * serial
     */
    private static final long serialVersionUID = 101L;

    // Private members ****************************************************

    /**
     * Logger
     */
    private final Logger logger = LoggerFactory.getLogger(DefaultDevice.class);

    /**
     * Device identifier
     */
    private final String id;

    private final String userAgent;

    private final boolean actualDeviceRoot;

    private final String deviceRootId;

    /**
     * CapabilitiesHolder instance used to get capabilities values
     */
    private final CapabilitiesHolder capabilityHolder;

    /**
     * MarkupResolver used to obtain MarkUp for device
     */
    private final MarkupResolver markupResolver;

    /**
     * It is not serialized
     */
    private transient MarkUp markUp;


    /**
     * Build a Device.
     *
     * @param modelDevice        The ModelDevice instance modelDevice of this {@link Device}.
     * @param deviceRootId       Ancestor id
     * @param capabilitiesHolder The CapabilitiesHolder used to get capabilities values.
     * @param markupResolver     The MarkupResolver used to obtain MarkUp for this Device.
     */
    protected DefaultDevice(ModelDevice modelDevice, String deviceRootId,
                         CapabilitiesHolder capabilitiesHolder, MarkupResolver markupResolver) {

        Validate.notNull(modelDevice, "The modelDevice device must be not null");
        Validate.notNull(capabilitiesHolder,
                "The capabilitiesHolder must be not null");
        Validate.notNull(markupResolver, "The markupResolver must be not null");

        this.id = modelDevice.getID();
        this.userAgent = modelDevice.getUserAgent();
        this.actualDeviceRoot = modelDevice.isActualDeviceRoot();
        this.deviceRootId = deviceRootId;

        this.capabilityHolder = capabilitiesHolder;
        this.markupResolver = markupResolver;

    }

    /**
     * {@inheritDoc}
     */
    public String getId() {
        return id;
    }

    /**
     * {@inheritDoc}
     */
    public String getUserAgent() {
        return userAgent;
    }

    /**
     * {@inheritDoc}
     */
    public String getCapability(String name)
            throws CapabilityNotDefinedException {
        return this.capabilityHolder.getCapabilityValue(name);
    }

    /**
     * {@inheritDoc}
     */
    public Map getCapabilities() {

        return new HashMap(capabilityHolder.getAllCapabilities());
    }

    /**
     * {@inheritDoc}
     */
    public MarkUp getMarkUp() {

        // waste memory?
        if (markUp == null) {
            logger.trace("Resolving MarkUp for device...");
            markUp = markupResolver.getMarkupForDevice(this);
        }
        return markUp;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isActualDeviceRoot() {
        return actualDeviceRoot;
    }

    /**
     * {@inheritDoc}
     *
     * @return id of ancestor device.
     */
    public String getDeviceRootId() {
        String ret = deviceRootId;
        //we choose to return an empty string, instead of 'generic'
        if (deviceRootId.equals(Constants.GENERIC)) ret = StringUtils.EMPTY_STRING;

        return ret;
    }

    // Commons methods ****************************************************

    /**
     * {@inheritDoc}
     * It assumes two DefaultDevices with equals root ModelDevices are equals.
     */
    public boolean equals(Object obj) {

        EqualsBuilder eb = new EqualsBuilder();

        eb.appendSuper(getClass().isInstance(obj));

        if (eb.isEquals()) {
            DefaultDevice other = (DefaultDevice) obj;
            eb.append(id, other.id);
        }

        return eb.isEquals();
    }

    public int hashCode() {

        return new HashCodeBuilder(63, 89).append(getClass()).append(id).toHashCode();
    }

    public String toString() {

        return new ToStringBuilder(this).append(id).toString();
    }

}
