/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core;

import net.sourceforge.wurfl.core.resource.ModelDevice;
import net.sourceforge.wurfl.core.resource.ModelDeviceWithAncestorId;
import net.sourceforge.wurfl.core.resource.WURFLModel;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.text.StrBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * DeviceProvider default implementation.
 * <p>
 * It create DefaultDevices using the CapabilitiesHolderFactory and
 * MarkupResolver. By default it uses DefaultCapabilitiesHolderFactory and
 * DefaultMarkupResolver. Several constructors permit to customize the building
 * procedure.
 * </p>
 *
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * @version $Id: DefaultDeviceProvider.java 668 2011-06-14 07:07:44Z ermanno.franco $
 */
public class DefaultDeviceProvider extends AbstractDeviceProvider {

    /**
     * Log
     */
    private final Logger logger = LoggerFactory.getLogger(DefaultDeviceProvider.class);

    /**
     * The MarkupResolver used to initialize DefaultDevice
     */
    private final MarkupResolver markupResolver;

    /**
     * The CapabilitiesHolderFactory used to obtain CapabilitiesHolder
     */
    private final CapabilitiesHolderFactory capabilitiesHolderFactory;

    // Constructors *******************************************************

    /**
     * Create DefaultDeviceProvider by WURFLModel, CapabilitiesHolderFactory and
     * MarkupResolver.
     *
     * @param model                     The WURFLModel used to obtain the root {@link ModelDevice}.
     * @param capabilitiesHolderFactory The CapabilitiesHolderFactory used to create
     *                                  CapabilitiesHolder for the devices.
     * @param markupResolver            The MarkupResolver used to initialize new devices.
     */
    public DefaultDeviceProvider(WURFLModel model,
                                 CapabilitiesHolderFactory capabilitiesHolderFactory,
                                 MarkupResolver markupResolver) {

        super(model);

        Validate.notNull(capabilitiesHolderFactory,
                "capabilitiesHolderFactory must be not null.");
        Validate.notNull(markupResolver, "markupResolver must be not null.");

        this.capabilitiesHolderFactory = capabilitiesHolderFactory;
        this.markupResolver = markupResolver;
    }

    /**
     * Create DefaultDeviceProvider by WURFLModel and MarkupResolver.
     * <p>
     * It uses DefaultMarkupResolver as MarkupResolver implementation.
     * </p>
     *
     * @param model The {@link WURFLModel} used to obtain the root
     *              {@link ModelDevice}.
     */
    public DefaultDeviceProvider(WURFLModel model,
                                 CapabilitiesHolderFactory capabilitiesHolderFactory) {

        this(model, capabilitiesHolderFactory, new DefaultMarkupResolver());
    }

    /**
     * Create DefaultDeviceProvider by WURFLModel and MarkupResolver.
     * <p/>
     * <p>
     * It uses DefaultCapabilitiesHolderFactory as CapabilitiesHolderFactory
     * implementation.
     * </p>
     *
     * @param model          The {@link WURFLModel} used to obtain the root
     *                       {@link ModelDevice}.
     * @param markupResolver The {@link MarkupResolver} used to initialize new devices.
     */
    public DefaultDeviceProvider(WURFLModel model, MarkupResolver markupResolver) {

        this(model, new DefaultCapabilitiesHolderFactory(model), markupResolver);
    }

    /**
     * Create DefaultDeviceProvider by WURFLModel.
     * <p/>
     * <p>
     * It uses DefaultMarkupResolver as MarkupResolver implementation and
     * DefaultCapabilitiesHolderFactory as CapabilitiesHolderFactory
     * implementation.
     * </p>
     *
     * @param model The WURFLModel used to obtain the root ModelDevice.
     */
    public DefaultDeviceProvider(WURFLModel model) {
        this(model, new DefaultCapabilitiesHolderFactory(model),
                new DefaultMarkupResolver());
    }

    // Business methods ***************************************************

    /**
     * {@inheritDoc}
     */
    public Device getDevice(String id) throws DeviceNotDefinedException {

        Validate.notEmpty(id, "The id must be not null String");

        // Load modelDevice
        ModelDeviceWithAncestorId fatModelDevice = getModelDeviceWithAncestorId(id);

        // Create Device
        assert fatModelDevice.getModelDevice() != null : "modelDevice is null";

        CapabilitiesHolder capabilityHolder = createCapabilitiesHolder(fatModelDevice.getModelDevice());

        Device device = new DefaultDevice(fatModelDevice.getModelDevice(), fatModelDevice.getAncestorId(), capabilityHolder,
                markupResolver);

        if (logger.isTraceEnabled()) {
            StrBuilder logBuilder = new StrBuilder();
            logBuilder.append("Return device: ").append(device);

            logger.trace(logBuilder.toString());
        }

        return device;

    }

    // support methods ****************************************************

    /**
     * Factory method to build CapabilityHolder instances.
     *
     * @param modelDevice The ModelDevices root of the creating device.
     */
    protected CapabilitiesHolder createCapabilitiesHolder(
            ModelDevice modelDevice) {

        return capabilitiesHolderFactory.create(modelDevice);
    }

}
