/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core;

import net.sourceforge.wurfl.core.cache.CacheProvider;
import net.sourceforge.wurfl.core.cache.LRUMapCacheProvider;
import net.sourceforge.wurfl.core.matchers.MatcherManager;
import net.sourceforge.wurfl.core.request.WURFLRequest;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.text.StrBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>
 * This is the default implementation of WURFLService. It is responsible to
 * match the request user-agent with a managed device's identifier and returns a
 * Device instance.
 * </p>
 * <p>
 * MatcherManager instance is used to match WURFLRequest with one ModelDevice
 * identifier.
 * </p>
 * <p>
 * It uses a CacheProvider to cache Devices instance by the WURFLRequests.
 * </p>
 * 
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * 
 * @version $Id: DefaultWURFLService.java 672 2011-06-22 19:22:21Z ermanno.franco $
 */
public class DefaultWURFLService implements WURFLService {

	/** Log */
	private final Logger logger = LoggerFactory.getLogger(DefaultWURFLService.class);

	/** Caches device build from WURFLRequest */
	private CacheProvider cacheProvider;

	/** Matches the WURFLRequest to obtain device identifier */
	private MatcherManager matcherManager;

	/** Builds devices */
	private DeviceProvider deviceProvider;
	
	// Constructors *******************************************************

	/**
	 * Build DefaultWURFLService from their private members.
	 * 
	 * @param matcherManager
	 *            The MatcherManager instance used to match the WURFLRequest to
	 *            handled user-agent.
	 * @param deviceProvider
	 *            The DeviceProvider instance used to create Device instances.
	 * @param cacheProvider
	 *            The CacheProvider instance used to cache created Devices.
	 */
	public DefaultWURFLService(MatcherManager matcherManager,
			DeviceProvider deviceProvider, CacheProvider cacheProvider) {

		this.matcherManager = matcherManager;
		this.deviceProvider = deviceProvider;
		this.cacheProvider = cacheProvider;

		logger.info("DefaultWURFLService created");
	}

	/**
	 * Build DefaultWURFLService from their private members using
	 * {@link LRUMapCacheProvider} by default.
	 * 
	 * @param matcherManager
	 *            The MatcherManager instance used to match the WURFLRequest to
	 *            handled user-agent.
	 * @param deviceProvider
	 *            The DeviceProvider instance used to create Device instances.
	 */
	public DefaultWURFLService(MatcherManager matcherManager,
			DeviceProvider deviceProvider) {

		this(matcherManager, deviceProvider, new LRUMapCacheProvider());
	}
			
	protected CacheProvider getCacheProvider() {
		return cacheProvider;
	}
	
	// Business methods ***************************************************

	/**
	 * {@inheritDoc}
	 */
	public Device getDeviceForRequest(WURFLRequest request) {

		// Check input
		Validate.notNull(request, "The request is null");

		// Get device from cache
		Object deviceKey = createDeviceKey(request);
		Device device = (Device) cacheProvider.get(deviceKey);

		if (logger.isDebugEnabled() && device != null) {
			logger.debug("Found cached device: " + device);
		}

		if (device == null) {

			if (logger.isDebugEnabled()) {
				logger.debug("Not found device cached with key: " + deviceKey
						+ ", matching...");
			}

			// Matching device
			String deviceId = matcherManager.matchRequest(request);
			device = deviceProvider.getDevice(deviceId);

			cacheProvider.put(deviceKey, device);

		} // if (device == null)

		assert device != null;

		return device;
	}

	// Support methods ****************************************************

	/**
	 * Create a key to store device identifier in cache. Actually it return the
	 * user-agent followed by uaprof.
	 * 
	 * <p>
	 * The String is a good choice for key because they cache the hashCode.
	 * </p>
	 * 
	 * @param request
	 *            The request to create key from.
	 * @return A Object used ad key to cache a Device by the corresponding
	 *         WURFLRequest.
	 */
	protected Object createDeviceKey(WURFLRequest request) {

		// Check preconditions
		assert request != null : "The request is null";

		/*
		 * Maybe useful to generate String or not? String are hashCode cached.
		 * so more fast
		 */
		return new StrBuilder(request.getUserAgent()).append(
				request.getUserAgentProfile()).toString();
	}

}
