/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core.cache;

import net.sourceforge.wurfl.core.utils.CollectionFactory;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;

/**
 * {@link CacheProvider} HashMap implementation.
 *
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * @version $Id: HashMapCacheProvider.java 649 2011-05-20 18:28:18Z ermanno.franco $
 */
public class HashMapCacheProvider implements CacheProvider {

    /**
     * The hashMap containing the cached objects
     */
    private java.util.Map cache;

    /**
     * Initial capacity of hashMap
     */
    private int initialCapacity = 6000;

    /**
     * load factor of hashmap
     */
    private float loadFactor = 0.75f;

    /**
     * load factor of hashmap
     */
    private int concurrentWrites = 16;

    /**
     * Logger
     */
    private final Logger logger = LoggerFactory.getLogger(HashMapCacheProvider.class);

    // Constructors *******************************************************

    /**
     * Default constructor
     */
    public HashMapCacheProvider() {
        this(60000);
    }

    /**
     * Constructor by the initial capacity.
     *
     * @param initialCapacity
     */
    public HashMapCacheProvider(int initialCapacity) {
        this(initialCapacity, .75f);
    }

    /**
     * Constructor by the initial capacity and load factor.
     *
     * @param initialCapacity The hashMap initial capacity.
     * @param loadFactor      The hashMap load factor.
     */
    public HashMapCacheProvider(int initialCapacity, float loadFactor) {
        this(initialCapacity, loadFactor, 16);
    }

    /**
     * Constructor by the initial capacity, loadFactor and maximum concurrent
     * writes
     *
     * @param initialCapacity  The hashMap initial capacity.
     * @param loadFactor       The hashMap load factor.
     * @param concurrentWrites The maximum thread can write the cache.
     */
    public HashMapCacheProvider(int initialCapacity, float loadFactor,
                                int concurrentWrites) {

        this.initialCapacity = initialCapacity;
        this.loadFactor = loadFactor;
        this.concurrentWrites = concurrentWrites;

        cache = createCache();

        if (logger.isInfoEnabled()) {
            StringBuffer logBuffer = new StringBuffer(
                    "Created HashMapCacheProvider with initial capacity: ");
            logBuffer.append(initialCapacity);
            logBuffer.append(" load factor: ");
            logBuffer.append(loadFactor);
            logBuffer.append(" concurrent writes: ");
            logBuffer.append(concurrentWrites);

            logger.info(logBuffer.toString());
        }
    }

    // Access methods *****************************************************

    /**
     * Return the initial capacity of the hashMap.
     *
     * @return the initial capacity of the hashMap.
     */
    public int getInitialCapacity() {
        return initialCapacity;
    }

    /**
     * Return the hashMap load factor.
     *
     * @return the hashMap load factor.
     */
    public float getLoadFactor() {
        return loadFactor;
    }

    /**
     * Return the maximum threads can write the cache.
     *
     * @return a int representing the maximum thread can write the cache.
     */
    public int getConcurrentWrites() {
        return concurrentWrites;
    }

    // CacheProvider implementation ***************************************

    /**
     * {@inheritDoc}
     */
    public void clear() {
        cache.clear();

        logger.debug("Cache erased");
    }

    /**
     * {@inheritDoc}
     */
    public Object get(Object key) {

        Validate.notNull(key, "The key is null");

        Object target = cache.get(key);

        if (logger.isDebugEnabled()) {
            if (target == null) {
                logger.debug("The key: " + key + " is not in cache");
            } else {
                logger.debug("Retrieving from cache object: " + target
                        + " by key: " + key);
            }
        }

        return target;
    }

    /**
     * {@inheritDoc}
     */
    public void put(Object key, Object target) {

        Validate.notNull(key, "The key is null");

        if (logger.isDebugEnabled()) {
            logger.debug("Storing object: " + target + " with key: " + key);
        }

        cache.put(key, target);
    }

    /**
     * Factory method to create cache Map. Subclass can override this method to
     * change the backing HashMap implementation.
     */
    protected Map createCache() {
        return CollectionFactory.createConcurrentHashMap(initialCapacity, loadFactor, concurrentWrites);
    }

    // Commons methods ****************************************************

    public String toString() {
        return new ToStringBuilder(this).append(cache).toString();
    }
}
