/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core.classifiers;

import org.apache.commons.lang.Validate;
import org.apache.commons.lang.builder.ToStringBuilder;

import java.util.*;

/**
 * Filtered device data.
 * 
 * <p>
 * This class is responsible for collect devices classified according to their
 * user-agent.
 * </p>
 * 
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * 
 * @version $Id: FilteredDevices.java 672 2011-06-22 19:22:21Z ermanno.franco $
 */
public class FilteredDevices {

	/** The map of user-agent and device identifier. */
	private SortedMap/*String,String*/ data = new TreeMap();

	/** Default constructor */
	public FilteredDevices() {
		// Empty
	}

	/**
	 * Return the collected user-agent <b>naturally ordered</b>.
	 * 
	 * @return A {@link SortedSet} containing collected user-agent.
	 */
	public SortedSet getUserAgents() {
		return new TreeSet(data.keySet());
	}

	/**
	 * Return the collected user-agent <b>naturally ordered</b>.
	 * 
	 * @return A {@link SortedSet} containing collected user-agent.
	 */
	public Set getDevices() {
		return new HashSet(data.values());
	}

	/**
	 * Return a device by collected user-agent.
	 * 
	 * @param userAgent
	 *            The user-agent corresponding to the device tried.
	 * @return The device identifier collected corresponding given user-agent,
	 *         if the given user-agent is not collected it returns null.
	 */
	public String getDevice(String userAgent) {
		
		Validate.notNull(userAgent, "The userAgent is empty");
		
		return (String) data.get(userAgent);
	}
	
	public boolean containsDevice(String deviceId) {
		
		Validate.notEmpty(deviceId, "The deviceId is empty");
		
		return data.values().contains(deviceId);
	}
	
	/**
	 * Store device identifier with corresponding user-agent.
	 * 
	 * @param userAgent
	 *            The user-agent associated to device identifier.
	 * @param deviceId
	 *            The device identifier to store.
	 */
	public FilteredDevices putDevice(String userAgent, String deviceId) {
		
		Validate.notNull(userAgent, "user-agent cannot be null");
		Validate.notEmpty(deviceId, "The deviceId is empty");
		
		FilteredDevices newFilteredDevices = new FilteredDevices();
		newFilteredDevices.data.putAll(data);
		newFilteredDevices.data.put(userAgent, deviceId);
		
		return newFilteredDevices;
	}
	
	/**
	 * FIXME: Document it
	 * @return a fresh FilteredDevices
	 */
	public FilteredDevices clear() {		
		return new FilteredDevices();
	}

	
	// Common methods *****************************************************

	public String toString() {
		return new ToStringBuilder(this).append(getDevices()).toString();
	}

}
