/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core.matchers;

import net.sourceforge.wurfl.core.Constants;
import net.sourceforge.wurfl.core.classifiers.FilteredDevices;
import net.sourceforge.wurfl.core.request.WURFLRequest;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * The {@link Matcher} chain-of-responsibility.
 * <p/>
 * <p>
 * This class implements the chain-of-responsibility pattern for the user-agent
 * {@link Matcher}. It is best practice the latest matcher can handle
 * everything and at least must returns <code>generic</code>
 * </p>
 *
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * @version $Id: MatcherChain.java 672 2011-06-22 19:22:21Z ermanno.franco $
 */
public class MatcherChain {

    /**
     * The {@link List} of {@link Matcher}.
     */
    private final List matchers = new ArrayList();

    /**
     * Log
     */
    private final Logger logger = LoggerFactory.getLogger(MatcherChain.class);

    /**
     * Add a matcher to the chain.
     *
     * @param matcher The {@link Matcher} to add.
     */
    public void add(Matcher matcher) {
        matchers.add(matcher);
    }

    /**
     * Return the list containing the user-agent matchers.
     *
     * @return A {@link List} of {@link Matcher}s.
     */
    public List getMatchers() {
        return Collections.unmodifiableList(matchers);
    }

    /**
     * Perform matching across {@link Matcher} chains.
     *
     * @param request The {@link WURFLRequest} to match.
     * @return The matched device identifier.
     */
    public String match(WURFLRequest request, FilteredDevices filteredDevices) {

        for (Iterator iterator = matchers.iterator(); iterator.hasNext(); ) {

            Matcher matcher = (Matcher) iterator.next();

            if (matcher.canHandle(request.getUserAgent())) {

                if (logger.isDebugEnabled()) {
                    logger.debug("Matcher: " + matcher + " handle request: "
                            + request);
                }

                String matchedDeviceId = matcher.match(request, filteredDevices);

                if (logger.isDebugEnabled()) {
                    logger.debug("Request: " + request + " match device: " + matchedDeviceId);
                }

                return matchedDeviceId;

            } // end if matcher.canHandle(request)

        }// end for


        // No matcher found to match the request
        if (logger.isWarnEnabled()) {
            logger.warn("No any matcher can handle the request: " + request
                    + ", returning generic device.");
        }

        return Constants.GENERIC;
    }

}
