/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core.matchers;

import net.sourceforge.wurfl.core.classifiers.FilterChain;
import net.sourceforge.wurfl.core.classifiers.FilterChainFactory;
import net.sourceforge.wurfl.core.classifiers.FilteredDevices;
import net.sourceforge.wurfl.core.request.WURFLRequest;
import net.sourceforge.wurfl.core.resource.ModelDevice;
import net.sourceforge.wurfl.core.resource.WURFLModel;

import java.util.Iterator;
import java.util.Set;

/**
 * This class is responsible to manage the matching process. It is a facade
 * class of {@link Matcher}s and {@link net.sourceforge.wurfl.core.classifiers.Filter}s.
 * 
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * 
 * @version $Id: MatcherManager.java 672 2011-06-22 19:22:21Z ermanno.franco $
 */
public class MatcherManager {

	/** The matcher chain-of-responsibility used to match user-agents */
	private final MatcherChain matcherChain;

	/** The Filter chain used to classify devices */
	private final FilterChain filterChain;


	// Constructors *******************************************************

	/**
	 * Build a MatcherManager by model, matcher chain and filter chain.
	 * 
	 * @param model
	 *            The WURFLModel used to get devices from.
	 * @param filterChain
	 *            The used FilterChain.
	 * @param matcherChain
	 *            The used MatcherChain.
	 */
	public MatcherManager(WURFLModel model, FilterChain filterChain,
			MatcherChain matcherChain) {

		this.filterChain = filterChain;
		this.matcherChain = matcherChain;

		addFilteredDevices(model.getAllDevices());
	}

	/**
	 * Build a MatcherManager by model.
	 * 
	 * @param model
	 *            The WURFLModel used to get devices from.
	 */
	public MatcherManager(WURFLModel model) {


		matcherChain = new MatcherChainFactory().create();

        filterChain = new FilterChainFactory(matcherChain).create();

		addFilteredDevices(model.getAllDevices());
	}

	// Business methods ***************************************************

	/**
	 * Match the request to obtain a device identifier.
	 * 
	 * @param request
	 *            The {@link WURFLRequest} to match.
	 * @return A valid device identifier, <code>generic</code> device at least.
	 */
	public String matchRequest(WURFLRequest request) {

		FilteredDevices filteredDevices = filterChain
				.getFilteredDevices(request.getUserAgent());

		return matcherChain.match(request, filteredDevices);

	}

	/**
	 * Add devices to be collected.
	 * 
	 * @param devices
	 *            Set of ModelDevice to be collected by matchers.
	 */
	private void addFilteredDevices(Set/* ModelDevice */devices) {

		for (Iterator iterator = devices.iterator(); iterator.hasNext();) {

			ModelDevice device = (ModelDevice) iterator.next();
			String userAgent = device.getUserAgent();
			String deviceId = device.getID();            
			filterChain.filter(userAgent, deviceId);
		}

	}

}
