/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core.resource;

import net.sourceforge.wurfl.core.Constants;
import org.apache.commons.collections.Predicate;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * This static class verifies the Model consistency
 *
 * @author Fantayeneh Asres Gizaw
 * @author Filippo De Luca
 * @version $Id: WURFLConsistencyVerifier.java 649 2011-05-20 18:28:18Z ermanno.franco $
 */
public final class WURFLConsistencyVerifier {

    /**
     * Log
     */
    private static final Logger logger = LoggerFactory
            .getLogger(WURFLConsistencyVerifier.class);


    private WURFLConsistencyVerifier() {
    }

    /**
     * Verify devices collection consistency.
     *
     * @param devices The ModelDevices to verify.
     * @throws WURFLConsistencyException If the given ModelDevices is not consistent.
     */
    public static void verify(ModelDevices devices)
            throws WURFLConsistencyException {

        Map userAgents = new HashMap();
        Set hierarchyVerifiedDevices = new HashSet();

        logger.debug("Verifing generic existence...");
        verifyGeneric(devices);

        Iterator devicesIterator = devices.iterator();
        while (devicesIterator.hasNext()) {

            ModelDevice device = (ModelDevice) devicesIterator.next();

            if (logger.isDebugEnabled()) {
                logger.debug("Verifing device: " + device.getID());
            }

            if (logger.isTraceEnabled()) {
                logger.trace("Verifing " + device.getID() + " userAgent...");
            }
            verifyUserAgent(device, userAgents);
            userAgents.put(device.getUserAgent(), device);

            if (logger.isTraceEnabled()) {
                logger.trace("Verifing " + device.getID() + " hierarchy...");
            }
            verifyHierarchy(device, devices, hierarchyVerifiedDevices);
            hierarchyVerifiedDevices.add(device.getID());

            if (logger.isTraceEnabled()) {
                logger.trace("Verifing " + device.getID() + " groups...");
            }
            verifyGroups(device, devices);

            if (logger.isTraceEnabled()) {
                logger.trace("Verifing " + device.getID() + " capabilities...");
            }

            verifyCapabilities(device, devices);
        }

    }

    private static void verifyGeneric(ModelDevices devices)
            throws WURFLConsistencyException {

        assert devices != null : "devices is null";

        if (!devices.containsId(Constants.GENERIC)) {
            throw new GenericNotDefinedException();
        }
    }

    private static void verifyUserAgent(ModelDevice device, Map userAgents)
            throws UserAgentConsistencyException {

        assert device != null : "device is null";
        assert userAgents != null : "userAgents is null";

        if (userAgents.containsKey(device.getUserAgent())) {

            ModelDevice definingDevice = (ModelDevice) userAgents.get(device
                    .getUserAgent());

            throw new UserAgentNotUniqueException(device,
                    device.getUserAgent(), definingDevice);

        }
    }

    private static void verifyHierarchy(ModelDevice device,
                                        ModelDevices devices, Set hierarchyVerifiedDevicesId)
            throws HierarchyConsistencyException {

        assert device != null : "device is null";
        assert devices != null : "devices is null";
        assert hierarchyVerifiedDevicesId != null : "hierarchyVerifiedDevicesId is null";

        List hierarchy = new ArrayList(10);

        String deviceId = device.getID();
        assert !StringUtils.isEmpty(deviceId);

        hierarchy.add(deviceId);

        while (!Constants.GENERIC.equals(deviceId)) {

            ModelDevice examineDevice = devices.getById(deviceId);

            String fallbackId = examineDevice.getFallBack();

            // Not-Null verified bt ResourceHandler
            assert !StringUtils.isEmpty(fallbackId);

            if (hierarchyVerifiedDevicesId.contains(fallbackId)) {
                // OK
                return;
            }

            if (!devices.containsId(fallbackId)) {
                throw new OrphanHierarchyException(hierarchy);
            }

            int hierarchyIndex = hierarchy.indexOf(fallbackId);
            if (hierarchyIndex != -1) {

                List circularHierarchy = new LinkedList(hierarchy.subList(
                        hierarchyIndex, hierarchy.size()));
                throw new CircularHierarchyException(circularHierarchy);
            }

            hierarchy.add(fallbackId);

            deviceId = fallbackId;

        }

    }

    private static void verifyGroups(ModelDevice device, ModelDevices devices)
            throws GroupConsistencyException {

        assert device != null : "device is null";
        assert devices != null : "devices is null";

        ModelDevice generic = devices.getById(Constants.GENERIC);

        Set deviceGroups = device.getGroups();
        Set genericGroups = generic.getGroups();

        for (Iterator gIt = deviceGroups.iterator(); gIt.hasNext(); ) {
            String group = (String) gIt.next();

            if (!genericGroups.contains(group)) {
                throw new InexistentGroupException(device, group);
            }

        }

    }

    private static void verifyCapabilities(ModelDevice device,
                                           ModelDevices devices) throws CapabilityConsistencyException {

        assert device != null : "device is null";
        assert devices != null : "devices is null";
        assert devices.containsId(Constants.GENERIC) : "device do not containing "
                + Constants.GENERIC;

        ModelDevice generic = devices.getById(Constants.GENERIC);

        Map genericCapabilities = generic.getCapabilities();
        Map deviceCapabilities = device.getCapabilities();

        for (Iterator dcnIt = deviceCapabilities.keySet().iterator(); dcnIt
                .hasNext(); ) {

            String capabilityName = (String) dcnIt.next();

            if (logger.isTraceEnabled()) {
                logger.trace("Verifing device: " + device.getID()
                        + " capability: " + capabilityName);
            }

            if (!genericCapabilities.containsKey(capabilityName)) {
                throw new InexistentCapabilityException(device, capabilityName);
            }

            if (!(device.getGroupForCapability(capabilityName).equals(generic
                    .getGroupForCapability(capabilityName)))) {

                throw new BadCapabilityGroupException(device, capabilityName,
                        device.getGroupForCapability(capabilityName), generic
                        .getGroupForCapability(capabilityName));
            }
        }

    }

    static class EqualsDeviceIdPredicate implements Predicate {

        private String id;

        public EqualsDeviceIdPredicate(String id) {
            this.id = id;
        }

        public boolean evaluate(Object object) {
            ModelDevice other = (ModelDevice) object;

            return id.equals(other.getID());
        }

    }

}
