/**
 * Copyright (c) 2011 ScientiaMobile Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */

package net.sourceforge.wurfl.spring;

import net.sourceforge.wurfl.core.CapabilityNotDefinedException;
import net.sourceforge.wurfl.core.DeviceNotDefinedException;
import net.sourceforge.wurfl.core.GroupNotDefinedException;
import net.sourceforge.wurfl.core.resource.*;
import net.sourceforge.wurfl.core.web.WurflWebConstants;
import org.apache.commons.lang.Validate;
import org.springframework.context.ResourceLoaderAware;
import org.springframework.core.io.Resource;
import org.springframework.core.io.ResourceLoader;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

/**
 * Helper Class to instantiate the WURFL Model via Spring.
 * Use this class as Spring bean definition only in case of advanced WURFL Engine customization, e.g. to change cache provider or
 * CapabilitiesHolderFactory. In these cases you must supply this bean to these objects constructors.<br/>
 * Finally, you supply this bean as "wurflModel" property in SpringWurflManager bean definition.<br/>
 * See wurfl-custom-ctx.xml in Helloworld examples to clarify that.
 */
public class SpringWURFLModel implements WURFLModel, ResourceLoaderAware, WurflWebConstants {
    private WURFLModel delegate;
    private boolean set;
    private WURFLResource wurfl = null;
    private List wurflResources = null;
    private ResourceLoader resourceLoader;

    public void setWurfl(Resource wurfl) {
        Validate.notNull(wurfl);
        this.wurfl = new SpringXMLResourceAdapter(wurfl);

    }

    public void setWurflPatch(Resource wurflPatch) {
        Validate.notNull(wurflPatch);
        setWurflPatches(new Resource[]{wurflPatch});
    }

    public void setWurflPatches(Resource[] wurflPatches) {
        Validate.notNull(wurflPatches);
        Validate.notEmpty(wurflPatches, "wurfl patches path cannot be empty");

        List wurflResources = new ArrayList();
        for (int i = 0; i < wurflPatches.length; i++) {
            SpringXMLResourceAdapter resourceAdapter = new SpringXMLResourceAdapter(wurflPatches[i]);
            wurflResources.add(resourceAdapter);
        }
        this.wurflResources = wurflResources;
    }


    private void initDelegate() {
        if (wurfl == null) {
            Resource resource = resourceLoader.getResource(WURFL_DEFAULT_LOCATION);
            wurfl = new SpringXMLResourceAdapter(resource);
        }
        if (wurflResources == null) {
            delegate = new DefaultWURFLModel(wurfl);
        } else {
            delegate = new DefaultWURFLModel(wurfl, new WURFLResources(wurflResources));
        }
        set = true;
    }

    public String getVersion() {
        if (!set) initDelegate();
        return delegate.getVersion();
    }

    public ModelDevice getDeviceById(String s) throws DeviceNotDefinedException {
        if (!set) initDelegate();
        return delegate.getDeviceById(s);
    }

    public boolean isDeviceDefined(String s) {
        if (!set) initDelegate();
        return delegate.isDeviceDefined(s);
    }

    public Set getAllDevices() {
        if (!set) initDelegate();
        return delegate.getAllDevices();
    }

    public Set getAllDevicesId() {
        if (!set) initDelegate();
        return delegate.getAllDevicesId();
    }

    public Set getDevices(Set s) throws DeviceNotDefinedException {
        if (!set) initDelegate();
        return delegate.getDevices(s);
    }

    public List getDeviceHierarchy(ModelDevice modelDevice) throws DeviceNotInModelException {
        if (!set) initDelegate();
        return delegate.getDeviceHierarchy(modelDevice);
    }

    public ModelDevice getDeviceFallback(ModelDevice modelDevice) throws DeviceNotInModelException {
        if (!set) initDelegate();
        return delegate.getDeviceFallback(modelDevice);
    }

    public ModelDevice getDeviceAncestor(ModelDevice modelDevice) throws DeviceNotInModelException {
        if (!set) initDelegate();
        return delegate.getDeviceAncestor(modelDevice);
    }

    public int size() {
        if (!set) initDelegate();
        return delegate.size();
    }

    public Set getAllCapabilities() {
        if (!set) initDelegate();
        return delegate.getAllCapabilities();
    }

    public boolean isCapabilityDefined(String s) {
        if (!set) initDelegate();
        return delegate.isCapabilityDefined(s);
    }

    public String getGroupByCapability(String s) throws CapabilityNotDefinedException {
        if (!set) initDelegate();
        return delegate.getGroupByCapability(s);
    }

    public ModelDevice getDeviceWhereCapabilityIsDefined(ModelDevice modelDevice, String s) throws DeviceNotInModelException, CapabilityNotDefinedException {
        if (!set) initDelegate();
        return delegate.getDeviceWhereCapabilityIsDefined(modelDevice, s);
    }

    public Set getAllGroups() {
        if (!set) initDelegate();
        return delegate.getAllGroups();
    }

    public boolean isGroupDefined(String s) {
        if (!set) initDelegate();
        return delegate.isGroupDefined(s);
    }

    public Set getCapabilitiesForGroup(String s) throws GroupNotDefinedException {
        if (!set) initDelegate();
        return delegate.getCapabilitiesForGroup(s);
    }

    public Set getRootDevicesIds() {
        return delegate.getRootDevicesIds();
    }


    public void setResourceLoader(ResourceLoader resourceLoader) {
        this.resourceLoader = resourceLoader;
    }
}
